
/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) 2008 Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */
package factory;

import java.io.PrintStream;

import util.PlinguaCoreException;

/**
 * An abstract class for a simulator
 * 
 * @author Ignacio Perez-Hurtado
 * 
 */
public abstract class Simulator {

	private Psystem psystem;
	private boolean time;
	private PrintStream infoChannel = System.out;

	/**
	 * @param psystem:
	 *            The P system to be simulated
	 */
	public Simulator(Psystem psystem) {
		if (psystem == null)
			throw new NullPointerException("Null P system");
		this.psystem = psystem;
	}

	/**
	 * @return The P system simulated
	 */
	public Psystem getPsystem() {
		return psystem;
	}

	/**
	 * @return A PrintStream to write information about the simulation proccess
	 *         in real-time By defect it is System.out
	 */
	public PrintStream getInfoChannel() {
		return infoChannel;
	}

	/**
	 * @param infoChannel:
	 *            The new PrintStream
	 */
	public void setInfoChannel(PrintStream infoChannel) {
		this.infoChannel = infoChannel;
	}

	/**
	 * Make one step of simulation and generate a new configuration
	 * 
	 * @return false if the current configuration is a halting one. True if a
	 *         new configuration was generated.
	 * @throws PlinguaCoreException
	 *             if occurs a semantic error
	 */
	public abstract boolean step() throws PlinguaCoreException;

	/**
	 * Make one alternate step of the last one.
	 * 
	 * @return False if there is not alternate configurations. True if a new
	 *         configuration was generated.
	 * @throws UnsupportedOperationException
	 */
	public abstract boolean alternateStep()
			throws UnsupportedOperationException;

	/**
	 * @return The number of possible next configurations in the next level of
	 *         the non-deterministic tree
	 * @throws UnsupportedOperationException
	 */
	public abstract int countAlternatives()
			throws UnsupportedOperationException;

	/**
	 * Make one step back.
	 * 
	 * @return False if the current configuration is the initial one. True if a
	 *         new configuration was generated.
	 * @throws UnsupportedOperationException
	 */
	public abstract boolean stepBack() throws UnsupportedOperationException;

	/**
	 * Reset the simulator to the initial configuration
	 */
	public abstract void reset();

	/**
	 * @return The current configuration
	 */
	public abstract Configuration getCurrentConfig();
	public abstract double getTime();
	public boolean isTime() {
		return time;
	}

	public void setTime(boolean time) {
		this.time = time;
	}

}
