/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) 2008 Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */

package factory;

/**
 * This is an Abstract Factory for variants of P systems
 * 
 * @author Ignacio Pérez-Hurtado
 * 
 */

public abstract class VariantFactory {

	private final String defaultInputParsers[] = { "P-Lingua", "XML" };

	private final String defaultOutputParsers[] = { "XML" };

	/**
	 * @return The name of the variant, for instance: Active Membranes,
	 *         Transtion P systems...
	 */
	public abstract String getVariantName();

	/**
	 * @return A new empty P System of this variant
	 */
	public abstract Psystem createPsystem();

	/**
	 * @param psystem:
	 *            The P system to be simulated
	 * @param stepBackAllowed:
	 *            It is true if it is permitted step backs
	 * @param alternateStepsAllowed:
	 *            It is true if it is permitted alternate steps
	 * @return A new simulator for this variant with a specific P system
	 */
	public abstract Simulator createSimulator(Psystem psystem,
			boolean stepBackAllowed, boolean alternateStepsAllowed);

	/**
	 * Create an input parser for this variant.
	 * 
	 * @param inputParserID:
	 *            A String with the ID of the input parser to be created
	 * @return A new input parser for this variant
	 */
	public abstract InputParser createInputParser(String inputParserID);

	/**
	 * Create an output parser for this variant
	 * 
	 * @param psystem:
	 *            The P system
	 * @param outputParserID:
	 *            A String with the ID of the output parser to be created
	 * @return A new output parser for this variant
	 */
	public abstract OutputParser createOutputParser(Psystem psystem,
			String outputParserID);

	/**
	 * @return An array of String with all of the valid input parsers for this
	 *         variant, for defect: "P-Lingua" and "XML"
	 */
	public String[] inputParsers() {
		return defaultInputParsers;
	}

	/**
	 * @return An array of String with all of the valid output parsers for this
	 *         variant, for defect: "XML"
	 */
	public String[] outputParsers() {
		return defaultOutputParsers;
	}

}
