/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */

package util;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Collection;

import org.jdom.Attribute;
import org.jdom.Element;

/**
 * A class for a membrane for cell-like P systems
 * 
 * @author Ignacio Perez-Hurtado
 * 
 */
public class CellLikeMembrane extends Membrane {

	private CellLikeMembrane parentMembrane = null;

	private List<CellLikeMembrane> childMembranes;

	private boolean disolved = false;

	private int id = 0;

	private int nextId = 0;

	private CellLikeMembrane skinMembrane = null;

	private MultiSet<String> auxMultiSet = null;

	private MultiSet<String> newMultiSet = null;

	private boolean onlyEvolutionAllowed = false;

	private Collection<RuleReference> rulesSelected;

	/**
	 * Create a new skin membrane
	 */
	public CellLikeMembrane() {
		this("", null);

	}

	/**
	 * @param label:
	 *            A string for the label
	 */
	public CellLikeMembrane(String label) {
		this(label, null);
	}

	/**
	 * @param parentMembrane:
	 *            The parent membrane
	 */
	public CellLikeMembrane(CellLikeMembrane parentMembrane) {
		this("", parentMembrane);
	}

	/**
	 * @param label:
	 *            A String
	 * @param parentMembrane:
	 *            The parent membrane or null if this membrane will be the skin
	 *            one
	 */
	public CellLikeMembrane(String label, CellLikeMembrane parentMembrane) {
		super(label);
		this.parentMembrane = parentMembrane;
		if (parentMembrane != null) {
			if (parentMembrane.isDisolved())
				throw new NullPointerException(
						"The parent membrane is disolved.");
			parentMembrane.childMembranes.add(this);
			skinMembrane = parentMembrane.skinMembrane;
			id = skinMembrane.nextId;
			skinMembrane.nextId++;

		} else {
			skinMembrane = this;
			id = 0;
			nextId = 1;
		}

		childMembranes = new ArrayList<CellLikeMembrane>();
	}

	/**
	 * @return: The automatic ID for the membrane, the skin membrane has ID 0
	 */
	public int getId() {
		return id;
	}

	/**
	 * @return: The set of rules selected for the next step
	 */
	public Collection<RuleReference> getRulesSelected() {
		return rulesSelected;
	}

	/**
	 * Set the rules for the next step
	 * 
	 * @param rulesSelected
	 */
	public void setRulesSelected(Collection<RuleReference> rulesSelected) {
		this.rulesSelected = rulesSelected;
	}

	/**
	 * An auxiliar multiset to make calcules, if it is reset, a new multiset
	 * will be created
	 * 
	 * @return The auxiliar multiset
	 */
	public MultiSet<String> getAuxMultiSet() {
		if (auxMultiSet == null)
			auxMultiSet = new HashMultiSet<String>(getMultiSet());
		return auxMultiSet;
	}

	/**
	 * Set the auxiliar multiset to null
	 */
	public void resetAuxMultiSet() {
		auxMultiSet = null;
	}

	/**
	 * Reset the new multiset
	 */
	public void resetNewMultiSet() {
		newMultiSet = null;
	}

	/**
	 * @return true if this membrane only can run evolution rules
	 */
	public boolean isOnlyEvolutionAllowed() {
		return onlyEvolutionAllowed;
	}

	/**
	 * Set the onlyEvolutionAllowed flag
	 * 
	 * @param onlyEvolutionAllowed
	 */
	public void setOnlyEvolutionAllowed(boolean onlyEvolutionAllowed) {
		this.onlyEvolutionAllowed = onlyEvolutionAllowed;
	}

	/**
	 * @return True if this membrane is the skin one.
	 */
	public boolean isSkinMembrane() {
		return parentMembrane == null && !isDisolved();
	}

	/**
	 * @return The parent membrane or null if it is the skin
	 */
	public CellLikeMembrane getParentMembrane() {
		return parentMembrane;
	}

	/**
	 * @return A String with a representation of the membrane in P-Lingua format
	 */
	public String getRepresentation() {
		String str = Membrane.getChargeSymbol(getCharge()) + "[";
		for (int i = 0; i < childMembranes.size(); i++) {
			str += childMembranes.get(i).toString();
			if (i + 1 < childMembranes.size())
				str += ", ";
		}
		if (childMembranes.size() != 0 && getMultiSet().size() != 0)
			str += ", ";
		if (getMultiSet().size() != 0)
			str += getMultiSet().toString();

		str += "]'" + getLabel();
		return str;
	}

	public String toString() {
		String str;
		if (isSkinMembrane())
			str = "Skin membrane ";
		else
			str = "Membrane ";
		str += "ID: " + getId() + ", ";
		str += "Label: " + getLabel() + ", ";
		str += "Charge: " + Membrane.getCompleteChargeSymbol(getCharge());
		return str;
	}

	/**
	 * @return An Iterator for child membranes
	 */
	public Iterator<CellLikeMembrane> childMembranesIterator() {
		return childMembranes.iterator();
	}

	/**
	 * @return The number of child membranes
	 */
	public int childMembranesCount() {

		return childMembranes.size();
	}

	public void disolution() throws SemanticsException {
		if (isSkinMembrane())
			throw new SemanticsException("Skin Membrane can not be disolved.");
		if (isDisolved())
			throw new NullPointerException("The membrane is disolved.");
		parentMembrane.childMembranes.remove(this);
		parentMembrane.getMultiSet().addAll(getMultiSet());
		Iterator<CellLikeMembrane> it = childMembranes.iterator();
		while (it.hasNext()) {
			CellLikeMembrane childMembrane = it.next();
			childMembrane.parentMembrane = parentMembrane;
			parentMembrane.childMembranes.add(childMembrane);
		}
		childMembranes.clear();
		parentMembrane = null;
		disolved = true;

	}

	public boolean isDisolved() {
		return disolved;
	}

	/**
	 * @return Get a Collection with this membrane and all of the internal ones
	 */
	public Collection<CellLikeMembrane> getAllMembranes() {
		List<List<CellLikeMembrane>> l = new ArrayList<List<CellLikeMembrane>>();
		getAllMembranesRec(l, 0);
		List<CellLikeMembrane> r = new ArrayList<CellLikeMembrane>();
		for (int i = l.size() - 1; i >= 0; i--)
			r.addAll(l.get(i));
		return r;
	}

	private void getAllMembranesRec(
			List<List<CellLikeMembrane>> membranesByDepth, int depth) {
		if (depth == membranesByDepth.size())
			membranesByDepth.add(new ArrayList<CellLikeMembrane>());
		List<CellLikeMembrane> l = membranesByDepth.get(depth);
		l.add(this);
		Iterator<CellLikeMembrane> it = childMembranesIterator();
		while (it.hasNext())
			it.next().getAllMembranesRec(membranesByDepth, depth + 1);
	}

	public CellLikeMembrane clone() {
		return divideRec(this, getParentMembrane());
	}

	public Membrane division() throws SemanticsException {
		if (isSkinMembrane())
			throw new SemanticsException("Skin Membrane can not be divided.");
		if (isDisolved())
			throw new NullPointerException("The membrane is disolved.");
		return divideRec(this, getParentMembrane());
	}

	private CellLikeMembrane divideRec(CellLikeMembrane membrane,
			CellLikeMembrane parentMembrane) {
		CellLikeMembrane m = new CellLikeMembrane(membrane.getLabel(),
				parentMembrane);
		m.setCharge(membrane.getCharge());
		m.getMultiSet().addAll(membrane.getMultiSet());
		Iterator<CellLikeMembrane> it = membrane.childMembranesIterator();
		while (it.hasNext())
			divideRec(it.next(), m);
		return m;
	}

	public Element getElement() {
		Element e = new Element("membrane");

		e.setAttribute(new Attribute("label", getLabel()));
		if (getCharge() > 0)
			e.setAttribute(new Attribute("charge", "+1"));
		else if (getCharge() < 0)
			e.setAttribute(new Attribute("charge", "-1"));
		else
			e.setAttribute(new Attribute("charge", "0"));
		Iterator<CellLikeMembrane> it = childMembranesIterator();
		while (it.hasNext()) {
			CellLikeMembrane m = it.next();
			Element e1 = m.getElement();
			e.addContent(e1);
		}

		return e;
	}

	/**
	 * Another auxiliar multiset to make calcules, if it is reset, a new
	 * multiset will be created
	 * 
	 * @return The multiset
	 */
	public MultiSet<String> getNewMultiSet() {
		if (newMultiSet == null)
			newMultiSet = new HashMultiSet<String>();
		return newMultiSet;
	}

}
