/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */
package util;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;

/**
 * A multiset that uses HashMap
 * 
 * @author Ignacio Perez-Hurtado
 * 
 * @param <E>
 *            The type of the elements of the multiset
 */
public class HashMultiSet<E> implements MultiSet<E> {

	private EntrySet entrySet = null;

	public HashMultiSet() {
		entrySet = new EntrySet();
	}

	/**
	 * @param objects:
	 *            An initial Collection for the multiset
	 */
	public HashMultiSet(Collection<? extends E> objects) {
		entrySet = new EntrySet();
		addAll(objects);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#add(java.lang.Object, int)
	 */
	public boolean add(E object, int multiplicity) {

		if (object == null)
			throw new NullPointerException();
		if (multiplicity < 1)
			throw new IllegalArgumentException();
		synchronized (entrySet.map) {
			int currentCount = count(object) + multiplicity;
			entrySet.map.put(object, currentCount);
			entrySet.size += multiplicity;
		}
		return true;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#addAll(java.util.Collection, int)
	 */
	public boolean addAll(Collection<? extends E> objects, int multiplicity) {
		if (objects == null)
			throw new NullPointerException();
		if (multiplicity < 1)
			throw new IllegalArgumentException();
		boolean r = false;
		Iterator<? extends E> it;
		if (objects instanceof MultiSet) {
			it = ((MultiSet<? extends E>) objects).entrySet().iterator();
			while (it.hasNext()) {
				E obj = it.next();
				int c = ((MultiSet<? extends E>) objects).count(obj);
				add(obj, c * multiplicity);
			}
			r = true;
		} else {
			it = objects.iterator();
			while (it.hasNext())
				add(it.next(), multiplicity);
			r = true;
		}

		return r;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#count(java.lang.Object)
	 */
	public int count(Object object) {

		if (object == null)
			throw new NullPointerException();
		if (!contains(object))
			return 0;
		else
			return entrySet.map.get(object).intValue();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#countSubSets(java.util.Collection)
	 */
	public int countSubSets(Collection<?> objects) {
		int min = 0;

		MultiSet multiSet;

		if (objects == null)
			throw new NullPointerException();

		if (objects instanceof MultiSet)
			multiSet = (MultiSet) objects;
		else
			multiSet = new HashMultiSet(objects);

		Iterator it = multiSet.entrySet().iterator();
		boolean first = true;
		boolean end = false;
		while (it.hasNext() && !end) {
			Object o = it.next();
			int m1 = count(o);
			int m2 = multiSet.count(o);
			int c = m1 / m2;
			if (first || c < min) {
				min = c;
				if (min == 0)
					end = true;
			}
			first = false;
		}

		return min;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#entrySet()
	 */
	public Set<E> entrySet() {
		return entrySet;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#add(java.lang.Object)
	 */
	public boolean add(E arg0) {
		return add(arg0, 1);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#addAll(java.util.Collection)
	 */
	public boolean addAll(Collection<? extends E> arg0) {

		return addAll(arg0, 1);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#clear()
	 */
	public void clear() {
		entrySet.clear();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#contains(java.lang.Object)
	 */
	public boolean contains(Object arg0) {
		return entrySet.contains(arg0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#containsAll(java.util.Collection)
	 */
	public boolean containsAll(Collection<?> arg0) {

		return entrySet.containsAll(arg0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#isEmpty()
	 */
	public boolean isEmpty() {

		return entrySet.isEmpty();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#iterator()
	 */
	public Iterator<E> iterator() {
		return new EntryIterator(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#remove(java.lang.Object, int)
	 */
	public boolean remove(Object arg0, int multiplicity) {
		if (arg0 == null)
			throw new NullPointerException();
		if (multiplicity < 1)
			throw new IllegalArgumentException();

		int n = count(arg0);
		if (n < multiplicity)
			return false;

		if (n == multiplicity)
			entrySet.remove(arg0);
		else {
			synchronized (entrySet) {
				E obj = (E) arg0;
				entrySet.map.put(obj, n - multiplicity);
				entrySet.size -= multiplicity;
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#remove(java.lang.Object)
	 */
	public boolean remove(Object arg0) {
		if (arg0 == null)
			throw new NullPointerException();
		if (!contains(arg0))
			return false;
		int n = count(arg0);
		if (n == 1)
			entrySet.remove(arg0);
		else {
			synchronized (entrySet) {
				E obj = (E) arg0;
				entrySet.map.put(obj, n - 1);
				entrySet.size--;
			}
		}
		return true;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#removeAll(java.util.Collection)
	 */
	public boolean removeAll(Collection<?> arg0) {
		return entrySet.removeAll(arg0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#retainAll(java.util.Collection)
	 */
	public boolean retainAll(Collection<?> arg0) {
		return entrySet.retainAll(arg0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#size()
	 */
	public int size() {
		return entrySet.size;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#toArray()
	 */
	public Object[] toArray() {
		Object[] objs = new Object[entrySet.size()];
		Iterator it = iterator();
		int i = 0;
		while (it.hasNext()) {
			objs[i] = it.next();
			i++;
		}
		return objs;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Collection#toArray(T[])
	 */
	public <T> T[] toArray(T[] arg0) {

		T[] aux;
		if (arg0 == null)
			throw new NullPointerException();
		if (arg0.length < entrySet.size())
			aux = (T[]) new Object[entrySet.size()];
		else if (arg0.length == entrySet.size())
			aux = arg0;
		else {
			aux = arg0;
			aux[entrySet.size()] = null;
		}

		Iterator it = iterator();
		int i = 0;
		while (it.hasNext()) {

			Object obj = it.next();
			if (aux[i].getClass().isAssignableFrom(obj.getClass()))
				aux[i] = (T) obj;
			else
				throw new ArrayStoreException();
			i++;
		}
		return aux;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.MultiSet#subtraction(java.util.Collection)
	 */
	public boolean subtraction(Collection<?> objects) {

		if (objects == null) {
			System.out.println("NULL POINTER");
			throw new NullPointerException();

		}
		int c = countSubSets(objects);
		if (c == 0)
			return false;
		MultiSet m;
		if (objects instanceof MultiSet)
			m = (MultiSet) objects;
		else
			m = new HashMultiSet(objects);
		Iterator it = m.entrySet().iterator();
		synchronized (entrySet) {
			while (it.hasNext()) {
				E obj = (E) it.next();
				int n = count(obj);
				int n1 = m.count(obj);
				if (n == n1)
					entrySet.map.remove(obj);
				else
					entrySet.map.put(obj, n - n1);
				entrySet.size -= n1;
			}
		}
		return true;

	}

	private class EntryIterator implements Iterator<E> {
		private Iterator<E> iterator = null;

		private int count = 0;

		private E obj = null;

		private HashMultiSet<E> m;

		public EntryIterator(HashMultiSet<E> m) {
			this.iterator = m.entrySet().iterator();
			this.m = m;
		}

		public boolean hasNext() {
			if (iterator.hasNext())
				return true;
			else {
				int max = m.count(obj);
				if (count < max)
					return true;
				else
					return false;
			}
		}

		public E next() {
			if (!hasNext())
				return null;

			if (count >= m.count(obj) || obj == null) {
				count = 0;
				obj = iterator.next();
			}
			count++;
			return obj;
		}

		public void remove() {
			throw new UnsupportedOperationException();

		}

	}

	public String toString() {

		if (isEmpty())
			return "#";
		String str = "";
		Iterator<E> it = entrySet.iterator();
		while (it.hasNext()) {
			E e = it.next();
			str += e.toString();
			int n = count(e);
			if (n > 1)
				str += "*" + n;
			if (it.hasNext())
				str += ", ";
		}
		return str;
	}

	private class EntrySet implements Set<E> {
		private Map<E, Integer> map = null;

		private int size = 0;

		public EntrySet() {
			map = new HashMap<E, Integer>();
		}

		public boolean add(E arg0) {
			throw new UnsupportedOperationException();
		}

		public boolean addAll(Collection<? extends E> arg0) {
			throw new UnsupportedOperationException();
		}

		public synchronized void clear() {
			map.clear();
			size = 0;
		}

		public boolean contains(Object arg0) {
			return map.containsKey(arg0);
		}

		public boolean containsAll(Collection<?> arg0) {
			return map.keySet().containsAll(arg0);
		}

		public boolean isEmpty() {
			return map.isEmpty();
		}

		public Iterator<E> iterator() {
			return map.keySet().iterator();
		}

		public synchronized boolean remove(Object arg0) {
			if (map.containsKey(arg0))
				size -= map.get(arg0).intValue();
			return map.remove(arg0) != null;
		}

		public synchronized boolean removeAll(Collection<?> arg0) {
			Iterator it;
			boolean r = false;
			if (arg0 instanceof MultiSet)
				it = ((MultiSet<?>) arg0).entrySet().iterator();
			else
				it = arg0.iterator();
			while (it.hasNext())
				r = r || remove(it.next());
			return r;
		}

		public synchronized boolean retainAll(Collection<?> arg0) {
			boolean r = false;
			Iterator<E> it = map.keySet().iterator();
			while (it.hasNext()) {
				E obj = it.next();
				int c = map.get(obj).intValue();
				if (!arg0.contains(obj)) {
					it.remove();
					size -= c;
					r = true;
				}
			}
			return r;
		}

		public int size() {
			return map.size();
		}

		public Object[] toArray() {
			return map.keySet().toArray();
		}

		public <T> T[] toArray(T[] arg0) {
			return map.keySet().toArray(arg0);
		}

		public String toString() {

			return map.keySet().toString();
		}

		@Override
		public int hashCode() {
			final int PRIME = 31;
			int result = 1;
			result = PRIME * result
					+ ((map == null) ? 0 : map.keySet().hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			final EntrySet other = (EntrySet) obj;
			if (map == null) {
				if (other.map != null)
					return false;
			} else if (!map.keySet().equals(other.map.keySet()))
				return false;
			return true;
		}

	}

	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result
				+ ((entrySet == null) ? 0 : entrySet.map.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final HashMultiSet other = (HashMultiSet) obj;
		if (entrySet == null) {
			if (other.entrySet != null)
				return false;
		} else if (!entrySet.map.equals(other.entrySet.map))
			return false;
		return true;
	}

}
