/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */
package variants.am;

import factory.Configuration;
import factory.Simulator;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import util.CellLikeMembrane;
import util.Membrane;
import util.Rule;
import util.RuleReference;
import util.PlinguaCoreException;
import util.RulesSet;
import variants.am.rules.AmDisolutionRule;
import variants.am.rules.AmDivisionRule;
import variants.am.rules.AmEvolutionRule;
import variants.am.rules.AmRule;
import variants.am.rules.AmSendInRule;
import variants.am.rules.AmSendOutRule;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;

import util.HashMultiSet;


public class AmSimulator extends Simulator {

	
	private AmConfiguration config;
	private List<AmConfiguration> list=null;
	private Collection<CellLikeMembrane> membranes=null;
	private boolean finished=false;
	
	private Date date=null;
	private double time=0;
	
	public AmSimulator(AmPsystem psystem)
	{
		super(psystem);
		config=(AmConfiguration)psystem.getFirstConfiguration();
		
		list = new ArrayList<AmConfiguration>();
		list.add((AmConfiguration)config.clone());
		
	}
	
	
	




	public boolean isFinished() {
		return finished;
	}







	@Override
	public boolean alternateStep() throws UnsupportedOperationException{
		throw new UnsupportedOperationException();
	}

	@Override
	public int countAlternatives() throws UnsupportedOperationException{
		throw new UnsupportedOperationException();
	}

	@Override
	public Configuration getCurrentConfig() {
		return config;
	}

	@Override
	public void reset() {
		if (isInitialConfig())
			return;
		getInfoChannel().println("--> RESET SIMULATION");
		config=(AmConfiguration)getPsystem().getFirstConfiguration();
		list.clear();
		list.add((AmConfiguration)config.clone());
		finished=false;
		date=null;
		time=0;
		try
		{
		microStepSelection();
		}catch(Exception ex){}

	}

	public boolean microStepSelection() throws PlinguaCoreException
	{
		if (finished)
			return false;
		int c;
		
		
		membranes = config.getSkinMembrane().getAllMembranes();
		
		init(membranes);
		c = selectRules(membranes);
		
		getInfoChannel().println();
		getInfoChannel().println("Configuration: "+config.getNumber());
		
		if (isTime())
		{
			if (date==null)
				date=new Date();
			Date newDate= new Date();
			time = newDate.getTime()-date.getTime();
			getInfoChannel().println("Time: "+time/1000+" s.");
		}
		if (c==0)
			getInfoChannel().println("Halt configuration (No rule can be selected to be\nexecuted in the next step)");
		getInfoChannel().println("***********************************************");
		
		if (c==0)
		{
			finished=true;
			return false;
		}
		return true;
	
		
	}
	
	public boolean microStepExecution() throws PlinguaCoreException
	{
		if (finished)
			return false;
		executeRules(membranes);
		
		config.setNumber(config.getNumber()+1);
		list.add((AmConfiguration)config.clone());
		
		return true;
	}
	
	public boolean step() throws PlinguaCoreException
	{
		boolean b;
		b=microStepSelection();
		if (b) b=microStepExecution();
		return b;
		
	}	
	private void printInfoMembrane(CellLikeMembrane m)
	{
		if (getInfoChannel()==null)
			return;
		if (m.isSkinMembrane())
			getInfoChannel().print("@@@ SKIN MEMBRANE ");
		else
			getInfoChannel().print("### MEMBRANE ");
		getInfoChannel().print("ID: "+m.getId()+", ");
		getInfoChannel().print("Label: "+m.getLabel()+", ");
		String ch=Membrane.getChargeSymbol(m.getCharge());
		if (ch.equals(""))
			ch="0";
		getInfoChannel().println("Charge: "+ch);
		getInfoChannel().print("    Multiset: ");
		getInfoChannel().println(m.getMultiSet());
		if (!m.isSkinMembrane())
			getInfoChannel().println("    Parent Membrane ID: "+m.getParentMembrane().getId());
		if (m.childMembranesCount()>0)
			getInfoChannel().println("    Internal membranes count: "+m.childMembranesCount());
		if (m.getRulesSelected().size()>0)
		{
			getInfoChannel().println("    Rules Selected:");
			Iterator <RuleReference>it= m.getRulesSelected().iterator();
			while(it.hasNext())
				getInfoChannel().println("    "+it.next());
		}
		getInfoChannel().println();
		if (m.isSkinMembrane() && !config.getEnvironment().isEmpty())
		{
			getInfoChannel().println("~~~ENVIRONMENT: "+config.getEnvironment());
			getInfoChannel().println();
		}
			
			
		
		
	}
	
	private void init(Collection<CellLikeMembrane>membranes) throws PlinguaCoreException
	{
	
		Iterator <CellLikeMembrane>it = membranes.iterator();
		while(it.hasNext())
		{
			CellLikeMembrane m = it.next();
			m.resetAuxMultiSet();
		
			m.setOnlyEvolutionAllowed(false);
			m.setRulesSelected(null);
		}
		
		
	}
	
	private void executeRules(Collection<CellLikeMembrane>membranes) throws PlinguaCoreException
	{
		Iterator<CellLikeMembrane>it = membranes.iterator();
		while(it.hasNext())
			it.next().resetNewMultiSet();
		it = membranes.iterator();
		while(it.hasNext())
		{
			CellLikeMembrane m = it.next();
			
			
			Iterator<RuleReference>it1 = m.getRulesSelected().iterator();
			while(it1.hasNext())
			{
				RuleReference ref = it1.next();
				AmRule r = (AmRule)ref.getRule();
					r.executeRule(m,ref.getCountApplications(),config.getEnvironment());
			}
		}
	}
	
	
	
	
	private int selectRulesByMembrane(RulesSet rulesSet,CellLikeMembrane m) throws PlinguaCoreException
	{
		int totalCount=0;
		
		Collection <Rule>rules = rulesSet.get(m.getLabel(),m.getCharge()); 
		LinkedList <RuleReference>ruleReferences=new LinkedList<RuleReference>();
	
		if (rules != null)
		{
			Iterator <Rule>it = rules.iterator();
			while(it.hasNext())
			{
				Rule r = it.next();
				int count=selectRule(r,m);
				if (count>0)
				{
					totalCount+=count;
					RuleReference ref = new RuleReference(r);
					ref.incCountApplications(count);
					if (r instanceof AmEvolutionRule)
						ruleReferences.addFirst(ref);
					else
						ruleReferences.addLast(ref);
				}
				
			}
		}
		m.setRulesSelected(ruleReferences);
		return totalCount;
	}
	
	
	
	private int selectRule(Rule r,CellLikeMembrane m)
	{
		int count=0;
		if (!m.getLabel().equals(r.getLabel()))
			return 0;
		if (m.getCharge()>0 && r.getCharge()<=0 ||
			m.getCharge()==0 && r.getCharge()!=0 ||
		   m.getCharge()<0 &&  r.getCharge()>=0)
			return 0;
		
		if (!(r instanceof AmEvolutionRule) &&
				m.isOnlyEvolutionAllowed())
			return 0;
		
		if (r instanceof AmEvolutionRule)
		{
			AmRule r1 = (AmRule)r;
			
			count=m.getAuxMultiSet().count(r1.getLeftHandObject());
			if (count>0)
				m.getAuxMultiSet().entrySet().remove(r1.getLeftHandObject());
		}
		else
		if ((r instanceof AmDisolutionRule ||
			 r instanceof AmDivisionRule ||
			 r instanceof AmSendInRule) &&
			 !m.isSkinMembrane())
		{
			AmRule r1 = (AmRule)r;
			if (r instanceof AmSendInRule)
			{
				CellLikeMembrane p= m.getParentMembrane();
				if (p.getAuxMultiSet().contains(r1.getLeftHandObject()))
				{
					count=1;
					p.getAuxMultiSet().remove(r1.getLeftHandObject());
					m.setOnlyEvolutionAllowed(true);
				}
			}
			else
			{
				if (m.getAuxMultiSet().contains(r1.getLeftHandObject()))
				{
					count=1;
					m.getAuxMultiSet().remove(r1.getLeftHandObject());
					m.setOnlyEvolutionAllowed(true);
				}
			}
			
		}
		else
		if (r instanceof AmSendOutRule)
		{
			AmRule r1 = (AmRule)r;
			if (m.getAuxMultiSet().contains(r1.getLeftHandObject()))
			{
				count=1;
				m.getAuxMultiSet().remove(r1.getLeftHandObject());
				m.setOnlyEvolutionAllowed(true);
			}
		}
		
		return count;
			
			
	}
	
	
	
	private int selectRules(Collection<CellLikeMembrane>membranes) throws PlinguaCoreException
	{
		int count=0;
		Iterator<CellLikeMembrane>it = membranes.iterator();
		while(it.hasNext())
		{
			CellLikeMembrane m = it.next();
			
			count+=selectRulesByMembrane(getPsystem().getRules(),m);
			printInfoMembrane(m);
		}
		return count;
	}
	public boolean isInitialConfig()
	{
		return list.size()==1;
	}

	@Override
	public boolean stepBack() throws UnsupportedOperationException{
		
		int n = list.size();
		if (n<2)
			return false;
		config=(AmConfiguration)list.get(n-2).clone();
		list.remove(n-1);
		getInfoChannel().println("--> STEP BACK (Configuration: "+config.getNumber()+")");
		finished=false;
		try
		{
		microStepSelection();
		}catch(Exception ex){}
		return true;
	}







	public double getTime() {
		return time;
	}
	
	
	
	
	

}
