/*
 * This file is part of the pLinguaCore API 
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 *  pLinguaCore is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with P-Lingua compiler.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  pLinguaCore 1.0 beta
 *  Copyright (c) Ignacio Perez-Hurtado 
 *  perezh@us.es
 *  Research Group On Natural Computing
 *  http://www.gcn.us.es
 */
package variants.am.pLinguaInputParser;

import java.util.LinkedList;

import util.Alphabet;
import util.CellLikeMembrane;
import util.MultiSet;
import variants.am.pLinguaInputParser.javaCC.PLinguaParser;
import variants.am.pLinguaInputParser.javaCC.ParseException;
import variants.am.pLinguaInputParser.javaCC.Token;
import variants.am.rules.AmRule;
import java.util.HashMap;
import variants.am.AmPsystem;


import java.io.*;

import util.HashMultiSet;
public class ParserInterface implements IParserInterface{

	private static ParserInterface singleton=null;
	
	private HashMap modules;
	private boolean started=false;
	private LinkedList stack;
	private AmPsystem model;
	private int depthLevel=0;
	private int verbosityLevel=3;
	private boolean error=false;
	private String outputFile=null;
	private String inputFile=null;
	private PrintStream errorChannel=System.out;
	private PrintStream warningChannel=System.out;
	private PrintStream infoChannel=System.out;
	private PrintStream msgChannel=System.out;
	private ParserInterface()
	{
		modules=new HashMap();
		stack=new LinkedList();
		model=new AmPsystem();
		
		writeMsg("\r\n",1);
		writeInfo("Compiling...",3);
		
	}
	
	public static ParserInterface getInstance()
	{
		if (singleton==null)
			singleton=new ParserInterface();
		return singleton;
	}
	
	public void reset()
	{
		modules.clear();
		stack.clear();
		model=new AmPsystem();
		depthLevel=0;
		started=false;
		error=false;
	}
	
	
	public boolean isError() {
		return error;
	}

	public String msgError(String msg,Token token)
	{
		String str=msg;
		str+=" at line "+token.beginLine;
		str+=", column "+token.beginColumn;
		str+=".";
		return str;
	}
	

	public PrintStream getMsgChannel() {
		return msgChannel;
	}

	public void setMsgChannel(PrintStream msgChannel) {
		this.msgChannel = msgChannel;
	}

	public PrintStream getErrorChannel() {
		return errorChannel;
	}

	public void setErrorChannel(PrintStream errorChannel) {
		this.errorChannel = errorChannel;
	}

	public PrintStream getInfoChannel() {
		return infoChannel;
	}

	public void setInfoChannel(PrintStream infoChannel) {
		this.infoChannel = infoChannel;
	}

	public PrintStream getWarningChannel() {
		return warningChannel;
	}

	public void setWarningChannel(PrintStream warningChannel) {
		this.warningChannel = warningChannel;
	}

	public int getVerbosityLevel() {
		return verbosityLevel;
	}

	public void setVerbosityLevel(int verbosityLevel) {
		this.verbosityLevel = verbosityLevel;
	}

	public void addModule(String name, Token token, LinkedList parameters) throws ParseException {
		if (name.toLowerCase().equals("main"))
			name=name.toLowerCase();
		if (modules.containsKey(name))
			throw new ParseException(msgError("Module '"+name+"' redefined",token));
		for(int i=0;i<parameters.size();i++)
		{
			for (int j=0;j<parameters.size();j++)
				if (parameters.get(i).equals(parameters.get(j)) && j!=i)
					throw new ParseException(msgError("Invalid parameters",token));
		}
		if (parameters.size()>0)
			writeInfo("Reading module '"+name+"' with parameters "+parameters,5);
		else
			writeInfo("Reading module '"+name+"'",5);
		modules.put(name, new Module(name,token,parameters));
	
	
		
	}
	public ModuleInstance getActualInstance()
	{
		if (stack.size()>0)
			return (ModuleInstance)stack.getFirst();
		else
			return null;
	}
	public int getDepthLevel() {
		return depthLevel;
	}

	public void addRule(AmRule rule, Token token) {
		if (!started())
			return;
		model.getRules().add(rule);
		
		writeInfo("Adding rule: "+rule,4);
			
		
	}
	
	

	public void begin() {
		if (!modules.containsKey("main"))
			writeWarning("Module 'main' has not been defined.");
		else
		{
			writeInfo("Generating P system...",3);
			started=true;
			try
			{
				callModule("main",new LinkedList(),null,false);
				if (outputFile!=null)
					writeInfo("Generating xml output file...",3);
				
			}
			catch(ParseException ex){writeError(ex.getMessage());}
		}
		
	}

	public void callModule(String moduleName, LinkedList paramValues, Token token,boolean interactive) throws ParseException{
		
		if (!started())
			return;
		if (!modules.containsKey(moduleName))
			throw new ParseException(msgError("Undefined module name '"+moduleName+"'",token));
		Module m = (Module)modules.get(moduleName);
		if (interactive && m.getParameters().size()>0)
		{
			paramValues.clear();
			writeMsg("\r\nModule '"+moduleName+"':\r\n",0);
			for(int i=0;i<m.getParameters().size();i++)
			{
				int value=readInteger("Enter an integer value for parameter '"+m.getParameters().get(i)+"': ");
				paramValues.add(i,new Integer(value));
			}
		}
		
		if (m.getParameters().size()!=paramValues.size())
			throw new ParseException(msgError("Invalid number of parameters",token));
		ModuleInstance instance = new ModuleInstance(m,paramValues);
		stack.addFirst(instance);
		if (paramValues.size()>0)
			writeInfo("Calling module '"+moduleName+"' for "+m.getParameters()+" = "+paramValues,5);
		else
			writeInfo("Calling module '"+moduleName+"'",5);
		Token returnToken=PLinguaParser.token;
		PLinguaParser.token=m.getToken();
		PLinguaParser.resetNtk();
		PLinguaParser.moduleBody();
		PLinguaParser.token=returnToken;
		PLinguaParser.resetNtk();
		stack.removeFirst();
	
	}

	public void decDepthLevel() {
		depthLevel--;
		
		
	}

	public Alphabet getAlphabet() {
		
		return model.getAlphabet();
	}

	public int getVariableValue(String name, Token token) throws ParseException{
		if (!started())
			return 0;
		ModuleInstance instance=(ModuleInstance)stack.getFirst();
	
		if (!instance.isDefined(name))
			throw new ParseException(msgError("Undefined variable '"+name+"'",token));
		Integer value=instance.getVariableValue(name);
		return value.intValue();
			
	}

	public void incDepthLevel() {
		
		depthLevel++;
	
		
	}

	public void setSkinMembrane(CellLikeMembrane skin, Token token) {
		if (!started())
			return;
		if (model.getSkinMembrane()!=null)
			writeWarning(msgError("Redefined skin membrane",token));
		model.setSkinMembrane(skin);
		writeInfo("Adding initial configuration: "+skin,4);
		
	}

	public boolean started() {
		
		return started && depthLevel==0;
	}
	
	public void writeInfo(String msg,int verbosityLevel) {
		if (this.verbosityLevel>=verbosityLevel)
			infoChannel.println("Info: "+msg);
	}
	
	public void writeMsg(String msg,int verbosityLevel)
	{
		if (this.verbosityLevel>=verbosityLevel)
		msgChannel.print(msg);
	}
	public void writeError(String error) {
		this.error=true;
		if (verbosityLevel>0)
		errorChannel.println("Error: "+error);
		
	}
	
	
	
	public int readInteger(String msg) {
		int a=0;
		BufferedReader buffer =new BufferedReader (new InputStreamReader(System.in));
		try
		{
			boolean end=false;
			while(!end)
			{
				try
				{
					writeMsg(msg,0);
					String str=buffer.readLine();
					a=Integer.parseInt(str);
					end=true;
				}
				catch(NumberFormatException ex)
				{
					writeMsg("Invalid integer value.\r\n",0);
				}
			}
			
		}
		catch(IOException ex)
		{
			writeError(ex.getMessage());
		}
		return a;
		
	}
	
	public void writeWarning(String warning) {
		if (verbosityLevel>1)
		warningChannel.println("Warning: "+warning);
		
	}

	public void multiSetAsig(String label, MultiSet<String> ms, boolean inc) {
		if (!started())
			return;
		if (!inc || !model.getInitialMultiSets().containsKey(label))
		{
			model.getInitialMultiSets().put(label, ms);
			writeInfo("Adding initial multiset (label "+label+"): "+ms.toString(),4);
		}
		else
		{
			MultiSet<String> m = model.getInitialMultiSets().get(label);
			
			writeInfo("Adding objects to initial multiset (label "+label+"): "+ms.toString(),4);
			m.addAll(ms);
			
			if (!model.getInputMultiSet().containsKey(label))
				model.getInputMultiSet().put(label, new HashMultiSet<String>());
			model.getInputMultiSet().get(label).addAll(ms);
		}
		
	}

	public void addAlphabetObject(String object) {
		if (started())
			model.getAlphabet().addObject(object);
		
	}

	public String getInputFile() {
		return inputFile;
	}

	public void setInputFile(String inputFile) {
		this.inputFile = inputFile;
	}

	public String getOutputFile() {
		return outputFile;
	}

	public void setOutputFile(String outputFile) {
		this.outputFile = outputFile;
	}
	
	public AmPsystem getModel()
	{
		return model;
	}
	
	
}
